"""My collection of useful functions and doodads."""

from __future__ import annotations

import collections
import numpy as np
import os
import sys
import typing

from tabulate import tabulate


def is_even(val):
    """Return ``True`` if even, ``False`` if odd."""
    return not val % 2


def revcumsum(a: np.ndarray, axis=None, dtype=None, out=None) -> np.ndarray:
    """Reverse cumulative summation.

    See np.cumsum for full docs.
    """
    return np.cumsum(a[::-1], axis, dtype, out)[::-1]


def filename_noext(path: str) -> str:
    """Get the name of a file without the extension.

    Parameters
    ----------
    path
        string containing the path to a file.

    Note that this only removes the final extension:
    >>> filename_noext('path/to/file.ext')
    'file'
    >>> filename_noext('path/to/archive.tar.gz')
    'archive.tar'
    """
    return os.path.splitext(os.path.basename(path))[0]


def all_same_sign(iterable: typing.Iterable) -> bool:
    """Check if all the elements in the iterable have the same sign.

    Returns true or false.

    Parameters
    ----------
    iterable
        An iterable item with numeric values.
    """
    v = (all(item >= 0 for item in iterable)
         or all(item < 0 for item in iterable))
    return v


def print_table(headers,
                data,
                datafmt='cols',
                tablefmt='pipe',
                stream=sys.stdout,
                **kwargs) -> str:
    """Print a neatly-formatted table.

    Parameters
    ----------
    headers:
        Headers of the table.

    data:
        List of lists to print.

    datafmt = 'cols':
        Order of data in ``data``. If 'cols' (default), each list in ``data`` is
        a column. If 'rows', each list in ``data`` is a row.

    tablefmt = 'pipe':
        Format descriptor. See ``tabulate.tabulate_formats`` for a list.

    stream = sys.stdout:
        Stream to print to. If stream==None, don't print.

    kwargs:
        Additional arguments to ``tabulate.tabulate``.
    """
    # Check input
    if datafmt.lower() == 'cols':
        if len(headers) != len(data):
            raise ValueError("Number of headers must equal number of columns.")

        lengths_of_cols = [len(col) for col in data]
        for length in lengths_of_cols:
            if lengths_of_cols[0] != length:
                raise ValueError("Columns must be of equal length.")

        tabular_data = np.array(data).transpose().tolist()
    elif datafmt.lower() == 'rows':
        tabular_data = data
    else:
        raise ValueError(f"Unrecognized data format: {datafmt}")

    tabulated = tabulate(tabular_data,
                         headers=headers,
                         tablefmt=tablefmt,
                         **kwargs)
    print(tabulated, file=stream)

    return tabulated


def recursive_update(d: dict, u: dict) -> dict:
    """Recursively update a nested dictionary.

    Parameters
    ----------
    d:
        ``dict`` to update.
    u:
        ``dict`` to use for updating.

    Example
    -------
    >>> d = {'a': {1: 2}}
    >>> u = {'a': {2: 1}}
    >>> recursive_update(d, u)
    {'a': {1: 2, 2: 1}}

    Compare:
    >>> d.update(u)
    >>> print(d)
    {'a': {2: 1}}

    Source: https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth#3233356
    """
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            d[k] = recursive_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class Color:
    """RGB-based color representation."""
    def __init__(self, rgb, cmyk=None, rgb1=False):
        """Create a new Color.

        Parameters
        ----------
        rgb : tuple
            RGB tuple representing the color, with integer values ranging from 0
            to 255. Values are cast to ``int`` during construction: rounding may
            occur.

        cmyk : tuple, optional
            CMYK tuple representing the color, with values ranging from 0 to
            100. (default: None)

        rgb1 : bool, optional
            If ``True``, the argument to ``rgb`` has floating point values
            ranging from 0 to 1. This is still stored as a tuple of ``int``s
            from 0 to 255: rounding will occur. (default: False)
        """
        # yapf: disable
        isrgb   = lambda tup: len(tup) == 3 and all([v >= 0 and v <= 255 for v in tup])
        isrgb1  = lambda tup: len(tup) == 3 and all([v >= 0 and v <=   1 for v in tup])
        iscmyk  = lambda tup: len(tup) == 4 and all([v >= 0 and v <= 100 for v in tup])
        iscmyk1 = lambda tup: len(tup) == 4 and all([v >= 0 and v <=   1 for v in tup])
        # yapf: enable

        if cmyk is not None and not iscmyk(cmyk):
            raise ValueError(f"Invalid CMYK tuple: {cmyk}")

        if rgb1:
            if not isrgb1(rgb):
                raise ValueError(f"Invalid RGB1 tuple: {rgb}")
            rgb = [int(v*255) for v in rgb]
        else:
            if not isrgb(rgb):
                raise ValueError(f"Invalid RGB tuple: {rgb}")
            rgb = [int(v) for v in rgb]

        self.RGB = tuple(rgb)
        self.CMYK = tuple(cmyk) if cmyk is not None else None

    @property
    def HEX(self) -> str:
        """Hex code representation of the color's RGB value.

        Example
        -------
        >>> smokey = Color(rgb=(88, 89, 91))
        >>> smokey.HEX
        '#58595b'
        """
        hexcodes = [f'{v:02x}' for v in self.RGB]
        return '#' + ''.join(hexcodes)

    @property
    def RGB1(self) -> typing.Tuple[float, float, float]:
        """RGB value of the color, normalized to 1."""
        return tuple(v/255 for v in self.RGB)

    def __repr__(self):
        return f"Color(rgb={self.RGB}, cmyk={self.CMYK})"

    @classmethod
    def fromTennesseePalette(cls, color_name) -> cls:
        """Select a color from the University of Tennessee Knoxville brand guide.

        Color reference: https://brand.utk.edu/colors/palettes/

        Based on Dr. Mark Denavit's MATLAB function ``TennesseeColorPalette``.
        See https://github.com/denavit/MATLAB-Library for more details.
        """
        try:
            color = cls._TENNESSEE_COLORS[color_name.lower()]
        except KeyError:
            raise ValueError(f"unrecognized color name {color_name!r}")

        return cls(rgb=color['rgb'], cmyk=color['cmyk'])

    # yapf: disable
    _TENNESSEE_COLORS = {
        'orange':      { 'rgb': (255, 130,   0), 'cmyk': (  0,  50, 100,   0) },
        'white':       { 'rgb': (255, 255, 255), 'cmyk': (  0,   0,   0,   0) },
        'smokey':      { 'rgb': ( 88,  89,  91), 'cmyk': (  0,   0,   0,  80) },
        'valley':      { 'rgb': (  0, 116, 111), 'cmyk': (100,  50,  65,   0) },
        'torch':       { 'rgb': (230,  89,  51), 'cmyk': (  0,  85, 100,   0) },
        'globe':       { 'rgb': (  0, 108, 147), 'cmyk': (100,  18,  10,  50) },
        'limestone':   { 'rgb': (240, 237, 227), 'cmyk': (  5,   5,  10,   0) },
        'river':       { 'rgb': ( 81, 124, 150), 'cmyk': ( 70,  40,  25,  10) },
        'leconte':     { 'rgb': (141,  32,  72), 'cmyk': ( 40, 100,  60,  30) },
        'regalia':     { 'rgb': (117,  74, 126), 'cmyk': ( 55, 100,  25,  25) },
        'sunsphere':   { 'rgb': (254, 213,  53), 'cmyk': (  0,  20,  90,   0) },
        'rock':        { 'rgb': (167, 169, 172), 'cmyk': (  0,   0,   0,  40) },
        'legacy':      { 'rgb': ( 87, 149, 132), 'cmyk': ( 65,  20,  50,  10) },
        'summitt':     { 'rgb': (185, 225, 226), 'cmyk': ( 25,   0,  10,   0) },
        'buckskin':    { 'rgb': (112,  85,  80), 'cmyk': ( 60,  70,  70,  15) },
        'energy':      { 'rgb': (238,  62, 128), 'cmyk': (  0,  90,  20,   0) },
        'switchgrass': { 'rgb': (171, 193, 120), 'cmyk': ( 25,   0,  80,  10) },
        'fountain':    { 'rgb': ( 33, 151, 169), 'cmyk': ( 75,  15,  25,  10) },
        'eureka':      { 'rgb': (235, 234, 100), 'cmyk': ( 10,   0,  75,   0) },
        'eureka!':     { 'rgb': (235, 234, 100), 'cmyk': ( 10,   0,  75,   0) },
    }
    # yapf: enable


# In case some of my code still uses the module level variable
_TENNESSEE_COLORS = Color._TENNESSEE_COLORS



import textwrap
import warnings


class Boxer():
    """Class for creating plaintext "boxes".

    A box is defined by several parameters, laid out so::

        |<------------------ width ------------------>|
        <first><------------ rule -------------><right>
        <left><lpad><------- text -------><rpad><right>
        <left><lpad><------- text -------><rpad><right>
        <left><lpad><------- text -------><rpad><right>
        <left><lpad><------- text -------><rpad><right>
        <left><------------- rule -------------><final>
    """
    def __init__(self,
                 left,
                 right,
                 rule,
                 pad=' ',
                 first=None,
                 final=None,
                 rpad=None,
                 width=80):
        """Create a new Boxer.

        Parameters
        ----------
        left : str
            The string to use for the left side of the box.
        right : str
            The string to use for the right side of the box.
        rule : str
            The string to use for the rule at the top and bottom of the box.
        pad : str, optional
            The padding to use between `left`/`right` and the text in the box.
            (default: ' ')
        first : str, optional
            Alternate string to use as the first left-hand side, e.g. for multi-
            line comment style boxes. (default: None)
        final : str, optional
            Alternate string to use for the end of the box, e.g. for multiline
            comment style boxes. (default: None)
        rpad : str, optional
            Alternate padding to use for the right-hand side of the box.
            (default: None)
        width : int, optional
            Default width in characters of created boxes. (default: 80)
        """
        self.left = left
        self.right = right
        self.rule = rule
        self.first = left if first is None else first
        self.final = right if final is None else final
        self.lpad = pad
        self.rpad = pad if rpad is None else rpad
        self.width = width

    def textwidth(self, width=None):
        """Return the available width in characters for the Boxer.

        Parameters
        ----------
        width : int, optional
            Total box width to calculate from.
        """
        if width is None:
            width = self.width
        # textwidth is from the box size, minus the two ends, minus two pads
        return (width - len(self.left) - len(self.right) - len(self.lpad) -
                len(self.rpad))

    def box(self, text, width=None, wrap=True, strip=True):
        """Box some text, returned as a joined string.

        Parameters
        ----------
        text : str
            Text to box.
        width : int, optional
            Total width of the box. (default: self.width)
        wrap : bool, optional
            If True, wrap long lines in `text`. If False, warnings will be
            issued for long lines but they will be left as-is, creating a spiky
            box. (default: True)
        strip : bool, optional
            If True, strip trailing whitespace from each line. (default: True)
        """
        return '\n'.join(self.boxsplit(text, width, wrap, strip))

    def boxsplit(self, text, width=None, wrap=True, strip=True):
        """Box some text, returned as a list of lines.

        Parameters
        ----------
        text : str
            Text to box.
        width : int, optional
            Total width of the box. (default: self.width)
        wrap : bool, optional
            If True, wrap long lines in `text`. If False, warnings will be
            issued for long lines but they will be left as-is, creating a spiky
            box. (default: True)
        strip : bool, optional
            If True, strip trailing whitespace from each line. (default: True)
        """
        if width is None:
            width = self.width

        textwidth = self.textwidth(width)
        toprulewidth = width - len(self.first) - len(self.right)
        bottomrulewidth = width - len(self.left) - len(self.final)

        # If `rule` is more than one character long, the logic for repeating it
        # is more complicated. We can fit it to the desired width via::
        #
        #   rule*(width // len(rule)) + rule[:width % len(rule)]
        #
        # The first will repeat `rule` until the next repetition would overfill
        # the width; the second takes only enough characters from `rule` to
        # fill out the rest.
        if len(self.rule) == 0:
            toprule = self.first
            bottomrule = self.final
        elif len(self.rule) == 1:
            toprule = self.first + self.rule*toprulewidth + self.right
            bottomrule = self.left + self.rule*bottomrulewidth + self.final
        else:
            # yapf: disable
            toprule = (
                self.first + self.rule*(toprulewidth // len(self.rule)) +
                self.rule[:toprulewidth % len(self.rule)] + self.right
            )
            bottomrule = (
                self.left + self.rule*(bottomrulewidth // len(self.rule)) +
                self.rule[:bottomrulewidth % len(self.rule)] + self.final
            )
            # yapf: enable

        lines = [toprule]
        text = text.splitlines()

        left = self.left + self.lpad
        right = self.rpad + self.right

        for i, line in enumerate(text):
            if wrap and len(line) > textwidth:
                wrappedlines = textwrap.wrap(line, width=textwidth)
            else:
                if len(line) > textwidth:
                    warnings.warn(f"box: line {i} exceeds box dimensions")
                wrappedlines = [line]

            for wline in wrappedlines:
                lines.append(left + wline.ljust(textwidth) + right)

        lines.append(bottomrule)
        if strip:
            lines = [l.rstrip() for l in lines]
        return lines


_CBoxer = Boxer(left=' *', right='* ', rule='*', first='/*', final='*/')
_PyBoxer = Boxer(left='#', right='#', rule='=')


def cbox(text, width=80, wrap=True):
    """Place text in a C-style multiline comment box.

    Parameters
    ----------
    text : str
        Text to box.
    width : int, optional
        Width of the box. (Note: not text width) (default: 80)
    wrap : bool, optional
        If True, wrap long lines in `text`. If False, warnings will be issued
        for long lines but they will be left as-is, creating a spiky box.
        (default: True)

    Example
    -------
    >>> print(cbox("hello world!", width=40))
    /**************************************
     * hello world!                       *
     **************************************/
    """
    return _CBoxer.box(text, width=width, wrap=wrap)


def pybox(text, width=80, wrap=True):
    """
    Parameters
    ----------
    text : str
        Text to box.
    width : int, optional
        Width of the box. (Note: not text width) (default: 80)
    wrap : bool, optional
        If True, wrap long lines in `text`. If False, warnings will be issued
        for long lines but they will be left as-is, creating a spiky box.
        (default: True)

    Example
    -------
    >>> print(pybox("hello world!", width=40))
    #======================================#
    # hello world!                         #
    #======================================#
    """
    return _PyBoxer.box(text, width=width, wrap=wrap)


def latex_name(shape, frac='nicefrac', join='-'):
    """Return LaTeX code for nicely typesetting a steel section name.

    Assumes the "by" part of the section is represented by an 'X', and that
    compound fractions are separated by '-' (hyphen, not endash).

    Only tested on W and HSS names so far.

    Parameters
    ----------
    shape : str
        Name of a steel section.
    frac : {'frac', 'tfrac', 'sfrac', 'nicefrac'}, optional
        The fraction macro to use. (default: 'nicefrac')
    join : str, optional
        The string that joins the integer and fractional parts of a compound
        fraction (default: '-')

    Example
    -------
    >>> name = 'HSS3-1/2X3-1/2X3/16'
    >>> latex_name(name)
    'HSS3-\\nicefrac{1}{2}$\\times$3-\\nicefrac{1}{2}$\\times$\\nicefrac{3}{16}'
    """
    recognized_macros = ['frac', 'tfrac', 'sfrac', 'nicefrac']
    if frac not in recognized_macros:
        raise ValueError(f'Unrecognized fraction macro {frac!r}')

    def frac_to_nicefrac(f):
        """Return LaTeX code for a nicefrac from a fraction like '3/16'. Does
        not support compound fractions."""
        (numer, denom) = f.split('/')
        return f"\\{frac}{{{numer}}}{{{denom}}}"

    shape_parts = shape.split('X')
    for [index, part] in enumerate(shape_parts):
        if '/' in part and '-' in part:  # need to activate compound fraction logic
            (integer, fraction) = part.split('-')
            newfraction = frac_to_nicefrac(fraction)
            shape_parts[index] = integer + join + newfraction
        elif '/' in part:  # need to activate fraction logic
            shape_parts[index] = frac_to_nicefrac(part)

    return '$\\times$'.join(shape_parts)



"""Email using mutt.

Mutt must be appropriately configured on the system and available on the PATH.
"""
import subprocess as sub


def send(to, subject, message, attach=None):
    """Send an email using mutt.
    
    Parameters
    ----------
    to : List[str]
        Email address(es) of the recipient(s).
    subject : str
        Subject line of the email.
    message : str
        The body of the message.
    attach : List[str], optional
        File(s) to attach. Can be a str if only one file.

    Returns
    -------
    process : subprocess.CompletedProcess
        Completed process information from running mutt.
    """
    if isinstance(to, str):
        to = [to]

    cmd = ['mutt', *to, '-s', subject]

    if attach is not None:
        if isinstance(attach, str):
            attach = [attach]
        cmd.extend(['-a', *attach, '--'])

    process = sub.run(cmd, input=message.encode())
    return process



from __future__ import annotations

import enum
import warnings

import numpy as np
from scipy.stats import lognorm
from scipy.optimize import fsolve
from scipy.interpolate import interp1d, interp2d

from . import asce7_16
from .units import process_unit_input


__all__ = [
    'acmrxx',
    'beta_total',
    'fundamental_period',
    'mapped_value',
    'seismic_response_coeff',
    'sf1',
    'smt',
    'ssf',
]


def _convert(value, units):
    return process_unit_input(value, units, convert=True).value


def acmrxx(beta_total, collapse_prob, xin=0.622):
    """Compute the acceptable value of the adjusted collapse margin ratio (ACMR).

    Parameters
    ----------
    beta_total:
        Total uncertainty present in the system
    collapse_prob:
        Collapse probability being checked (e.g. 0.20 for ACMR20)
    xin = 0.622:
        Starting value for the nonlinear solution. Tweak this if there are
        convergence issues.

    Ref: FEMA P695 Section 7.4
    """

    # Solve lognorm.cdf as a substitute for MATLAB's logninv function
    def f(x):
        return lognorm.cdf(x, beta_total) - collapse_prob

    X = fsolve(f, xin)

    return 1/X[0]


# Uncertainty values for each rating
_rating_values = {
    'A': 0.10,
    'B': 0.20,
    'C': 0.35,
    'D': 0.50,
}


def beta_total(rating_DR: str,
               rating_TD: str,
               rating_MDL: str,
               mu_T: float = 3.0) -> float:
    """Compute the total uncertainty present in th system.

    Parameters
    ----------
    rating_DR:
        Rating of the design requirements for the system
    rating_TD:
        Rating of the test data for the system
    rating_MDL:
        Rating of the model's representation of the system
    mu_T:
        Period-based ductility

    Ref: FEMA P695 Section 7.3.1
    """
    beta_DR = _rating_values[rating_DR.upper()]
    beta_TD = _rating_values[rating_TD.upper()]
    beta_MDL = _rating_values[rating_MDL.upper()]
    beta_RTR = min((0.1 + 0.1*mu_T, 0.4))
    beta = np.sqrt(beta_RTR**2 + beta_DR**2 + beta_TD**2 + beta_MDL**2)

    return round(beta*40)/40


_mapped_value_dict = {
    "dmax": {
        "ss": 1.5,
        # Actually 0.60 but "should be taken as less than 0.60" *eyeroll*
        "s1": 0.59999999999,
        "fa": 1.0,
        "fv": 1.50,
        "sms": 1.50,
        "sm1": 0.90,
        "sds": 1.0,
        "sd1": 0.60,
        "ts": 0.60,
    },
    "dmin": {
        "ss": 0.55,
        "s1": 0.132,
        "fa": 1.36,
        "fv": 2.28,
        "sms": 0.75,
        "sm1": 0.30,
        "sds": 0.50,
        "sd1": 0.20,
        "ts": 0.40,
    },
    "cmin": {
        "ss": 0.33,
        "s1": 0.083,
        "fa": 1.53,
        "fv": 2.4,
        "sms": 0.50,
        "sm1": 0.20,
        "sds": 0.33,
        "sd1": 0.133,
        "ts": 0.40,
    },
    "bmin": {
        "ss": 0.156,
        "s1": 0.042,
        "fa": 1.6,
        "fv": 2.4,
        "sms": 0.25,
        "sm1": 0.10,
        "sds": 0.167,
        "sd1": 0.067,
        "ts": 0.40,
    }
}
_mapped_value_dict["cmax"] = _mapped_value_dict["dmin"]
_mapped_value_dict["bmax"] = _mapped_value_dict["cmin"]


def mapped_value(value: str, sdc: str):
    """Retrieve the mapped seismic parameter for a given design category.

    Parameters
    ----------
    value:
        Mapped parameter to retrieve (SS, S1, Fa, Fv, SMS, SM1, SDS, SD1, Ts)
    """
    return _mapped_value_dict[sdc.lower()][value.lower()]


_T_INTERP = [
    0.25, 0.30, 0.35, 0.40, 0.45, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.2, 1.4, 1.6,
    1.8, 2.0, 2.2, 2.4, 2.6, 2.8, 3.0, 3.5, 4.0, 4.5, 5.0
]
_SNRT_INTERP = [
    0.785, 0.781, 0.767, 0.754, 0.755, 0.742, 0.607, 0.541, 0.453, 0.402, 0.350,
    0.303, 0.258, 0.210, 0.169, 0.149, 0.134, 0.119, 0.106, 0.092, 0.081, 0.063,
    0.053, 0.046, 0.041
]


def sf1(T, sdc):
    """Calculate scale factor 1, which scales ground motions to the MCE.

    Parameters
    ----------
    T : float
        Period of the structure

    sdc : str
        Seismic design category


    Ref: FEMA P695 Section
    """
    T = _convert(T, 's')
    if T <= _T_INTERP[0] or T >= _T_INTERP[-1]:
        raise ValueError(f"Period is out of range: T = {T}")

    f = interp1d(_T_INTERP, _SNRT_INTERP)
    SNRT = f(T)
    SMT = smt(T, sdc)

    return SMT/SNRT


def smt(T, sdc):
    """Calculate the SMT.

    Parameters
    ----------
    T : float
        Fundamental period of the structure (seconds)

    sdc : str
        Seismic design category
    """
    T = _convert(T, 's')
    SM1 = mapped_value("SM1", sdc)
    SMS = mapped_value("SMS", sdc)
    if T <= SM1/SMS:
        return SMS
    else:
        return SM1/T


#-------------------------------------------
# Spectral shape factor stuff
#-------------------------------------------
_Z_SSF_DICT = {
    "Dmax":
        np.array([
            [1.00, 1.05, 1.10, 1.13, 1.18, 1.22, 1.28, 1.33],
            [1.00, 1.05, 1.11, 1.14, 1.20, 1.24, 1.30, 1.36],
            [1.00, 1.06, 1.11, 1.15, 1.21, 1.25, 1.32, 1.38],
            [1.00, 1.06, 1.12, 1.16, 1.22, 1.27, 1.35, 1.41],
            [1.00, 1.06, 1.13, 1.17, 1.24, 1.29, 1.37, 1.44],
            [1.00, 1.07, 1.13, 1.18, 1.25, 1.31, 1.39, 1.46],
            [1.00, 1.07, 1.14, 1.19, 1.27, 1.32, 1.41, 1.49],
            [1.00, 1.07, 1.15, 1.20, 1.28, 1.34, 1.44, 1.52],
            [1.00, 1.08, 1.16, 1.21, 1.29, 1.36, 1.46, 1.55],
            [1.00, 1.08, 1.16, 1.22, 1.31, 1.38, 1.49, 1.58],
            [1.00, 1.08, 1.17, 1.23, 1.32, 1.40, 1.51, 1.61],
        ]),
    "Dmin":
        np.array([
            [1.00, 1.02, 1.04, 1.06, 1.08, 1.09, 1.12, 1.14],
            [1.00, 1.02, 1.05, 1.07, 1.09, 1.11, 1.13, 1.16],
            [1.00, 1.03, 1.06, 1.08, 1.10, 1.12, 1.15, 1.18],
            [1.00, 1.03, 1.06, 1.08, 1.11, 1.14, 1.17, 1.20],
            [1.00, 1.03, 1.07, 1.09, 1.13, 1.15, 1.19, 1.22],
            [1.00, 1.04, 1.08, 1.10, 1.14, 1.17, 1.21, 1.25],
            [1.00, 1.04, 1.08, 1.11, 1.15, 1.18, 1.23, 1.27],
            [1.00, 1.04, 1.09, 1.12, 1.17, 1.20, 1.25, 1.30],
            [1.00, 1.05, 1.10, 1.13, 1.18, 1.22, 1.27, 1.32],
            [1.00, 1.05, 1.10, 1.14, 1.19, 1.23, 1.30, 1.35],
            [1.00, 1.05, 1.11, 1.15, 1.21, 1.25, 1.32, 1.37],
        ])
}
_Z_SSF_DICT["Cmax"] = _Z_SSF_DICT["Dmin"]
_Z_SSF_DICT["Cmin"] = _Z_SSF_DICT["Dmin"]
_Z_SSF_DICT["Bmax"] = _Z_SSF_DICT["Dmin"]
_Z_SSF_DICT["Bmin"] = _Z_SSF_DICT["Dmin"]

_X_MU_T = [1.0, 1.1, 1.5, 2, 3, 4, 6, 8]
_Y_T = [0.5, 0.6, 0.7, 0.8, 0.9, 1.0, 1.1, 1.2, 1.3, 1.4, 1.5]


def ssf(T, mu_t, sdc):
    """Compute the spectral shape factor (SSF).

    Parameters
    ----------
    T:
        Period of the structure
    mu_t:
        Period-based ductility
    sdc:
        Seismic design category

    Ref: FEMA P695 Section
    """
    T = _convert(T, 's')
    try:
        Z_SSF = _Z_SSF_DICT[sdc]
    except KeyError:
        raise ValueError(f"Unknown seismic design category: {sdc}")

    if mu_t < 1:
        raise ValueError("mu_t must be greater than or equal to 1")

    if T <= 0.5:
        if mu_t >= 8:
            ssf = Z_SSF[0, -1]
        else:
            f = interp1d(_X_MU_T, Z_SSF[0, :])
            ssf = f(mu_t)[0]
    elif T >= 1.5:
        if mu_t >= 8:
            ssf = Z_SSF[-1, -1]
        else:
            f = interp1d(_X_MU_T, Z_SSF[-1, :])
            ssf = f(mu_t)[0]
    else:
        if mu_t >= 8:
            f = interp1d(_Y_T, Z_SSF[:, -1])
            ssf = f(mu_t)[0]
        else:
            f = interp2d(_X_MU_T, _Y_T, Z_SSF)
            ssf = f(mu_t, T)[0]

    return ssf


def fundamental_period(hn, Ct, x, sdc):
    """Calculate the fundamental period of the structure.

    Uses the following equation:

    .. math:: T = \\max(C_u C_t h_n^x, 0.25)

    Where :math:`C_u` is the coefficient on the upper limit of the calculated
    period from ASCE 7.

    Parameters
    ----------
    hn : float
        Height of the structure (ft)

    Ct : float
        Building period coefficient.

    x : float
        Exponent on building height.

    sdc : str
        Seismic design category.

    Returns
    -------
    T : float
        Fundamental period in seconds.
    """
    hn = _convert(hn, 'ft')
    SD1 = mapped_value('SD1', sdc)
    Ta = asce7_16.seismic.approximate_period(hn, Ct, x)
    Cu = asce7_16.seismic.period_upper_limit_coeff(SD1)

    return np.maximum(Cu*Ta, 0.25)


def seismic_response_coeff(R, T, sdc):
    """Calculate the seismic response coefficient, C_s.

    Parameters
    ----------
    R : float
        Response modification factor.

    T : float
        Fundamental period (s).

    sdc : str
        Seismic design category (Dmax, Dmin, etc.).

    Note that this function follows the assumptions and restrictions enforced by
    FEMA P695; namely, it is used only with mapped values from the
    ``mapped_values`` function, and for structures with periods of 4.0 s or
    lower. For a more general function, see ``asce7_16.seismic_response_coeff``.
    """
    T = _convert(T, 's')
    if T > 4.0:
        warnings.warn('seismic_response_coeff: Given period '
                      f'(T = {T} s) is out of FEMA P695 range')

    Ts = mapped_value('Ts', sdc)
    SD1 = mapped_value('SD1', sdc)
    SDS = mapped_value('SDS', sdc)

    if T <= Ts:
        Cs = SDS/R
    else:
        Cs = max(SD1/(T*R), 0.044*SDS)

    Cs = max(Cs, 0.01)

    return Cs



import unyt

if not hasattr(unyt, 'g0'):
    unyt.define_unit('g0', unyt.standard_gravity, R'\rm{g_0}')


def process_unit_input(in_, default_units=None, convert=False,
                       registry=None) -> unyt.unyt_array:
    """Process an input value that may or may not have units.

    Accepts the following input styles::

        in_ = 1000           ->  out_ = 1000*default_units
        in_ = (1000, 'psi')  ->  out_ = 1000*psi
        in_ = 1000*psi       ->  out_ = 1000*psi

    If `convert` is True, then values that come in with units are converted to
    `default_units` when returned::

        in_ = 1000           ->  out_ = 1000*default_units
        in_ = (1000, 'psi')  ->  out_ = (1000*psi).to(default_units)
        in_ = 1000*psi       ->  out_ = (1000*psi).to(default_units)

    Parameters
    ----------
    in_
        Input values.
    default_units : str, unyt.Unit, optional
        Default units to use if inputs don't have units associated already.
    convert : bool, optional
        Convert all inputs to `default_units` (default: False)
    registry : unyt.UnitRegistry, optional
        Necessary if the desired units are not in the default unit registry.
        Used to construct the returned unyt.unyt_array object.

    Returns
    -------
    q : unyt.unyt_array
    """
    if isinstance(in_, unyt.unyt_array):
        q = in_
    elif isinstance(in_, tuple):
        if len(in_) == 2:
            value, units = in_
            q = unyt.unyt_array(value, units, registry=registry)
        else:
            raise ValueError('Input tuple must be length 2; '
                             f'given had length {len(in_)}')
    else:
        q = unyt.unyt_array(in_, default_units, registry=registry)

    return q.to(default_units) if convert else q


def convert(value, units, registry=None):
    """Convert an input value to the given units, and return a bare quantity.

    If the input value doesn't have units, assumes the input is in the requested
    units already.
    
    Parameters
    ----------
    value : array_like
    units : str, unyt.Unit
    registry : unyt.UnitRegistry, optional

    Returns
    -------
    np.ndarray
    
    Examples
    --------
    >>> convert(30, 's')
    array(30.)
    >>> convert(30*ft, 'm')
    array(9.144)
    >>> convert(([24, 36, 48], 'inch'), 'furlong')
    array([0.0030303 , 0.00454545, 0.00606061])
    """
    return process_unit_input(value, units, convert=True, registry=registry).v



"""Seismic load requirements according to ASCE 7-16."""

import json

import numpy as np
import requests


def approximate_period(hn: float, Ct: float, x: float) -> float:
    """Calculate the approximate fundamental period T_a (sec).

    Parameters
    ----------
    hn:
        Structural height (ft, m).
    Ct:
        Building period coefficient. See Table 12.8-2.
    x:
        Exponent on structural height. See Table 12.8-2.

    Reference: Equation 12.8-7, Table 12.8-2
    """
    return Ct*hn**x


def period_upper_limit_coeff(SD1: float) -> float:
    """Calculate the coefficient for upper limit on calculated period, C_u.

    Parameters
    ----------
    SD1:
        Design spectral response acceleration parameter at 1 second (g).

    Reference: Table 12.8-1
    """
    return np.interp(SD1, [0.1, 0.15, 0.2, 0.3, 0.4], [1.7, 1.6, 1.5, 1.4, 1.4])


def seismic_response_coeff(R, Ie, SDS, SD1, S1, T, T_L) -> float:
    """Calculate the seismic response coefficient, C_s.

    Parameters
    ----------
    R:
        Response modification factor
    Ie:
        Seismic importance factor
    SDS:
        Design spectral response acceleration parameter at short periods (g).
    SD1:
        Design spectral response acceleration parameter at 1 second (g).
    S1:
        Mapped MCE spectral response acceleration parameter at 1 second (g).
    T:
        Building fundamental period (sec).
    T_L:
        Long-period transition period (sec).

    Reference: Section 12.8.1.1
    """
    R_Ie = R/Ie
    Cs_basic = SDS/R_Ie  # Eq. 12.8-2

    if T <= T_L:
        Cs_max = SD1/(T*R_Ie)  # Eq. 12.8-3
    else:
        Cs_max = SD1*T_L/(T**2*R_Ie)  # Eq. 12.8-4

    Cs_min = max(0.044*SDS*Ie, 0.01)  # Eq. 12.8-5
    if S1 >= 0.6:
        Cs_min = max(Cs_min, 0.5*S1/R_Ie)  # Eq. 12.8-6

    Cs = max(min(Cs_basic, Cs_max), Cs_min)
    return Cs


class SiteSpecificParameters():
    def __init__(self, pgauh, pgad, pga, fpga, pgam, ssrt, crs, ssuh, ssd, ss, fa,
                 sms, sds, sdcs, s1rt, cr1, s1uh, s1d, s1, fv, sm1, sd1, sdc1, sdc,
                 tl, cv, twoPeriodDesignSpectrum, twoPeriodMCErSpectrum,
                 verticalDesignSpectrum, verticalMCErSpectrum):
        self._pgauh: float = pgauh
        self._pgad: float = pgad
        self._pga: float = pga
        self._fpga: float = fpga
        self._pgam: float = pgam
        self._ssrt: float = ssrt
        self._crs: float = crs
        self._ssuh: float = ssuh
        self._ssd: float = ssd
        self._ss: float = ss
        self._fa: float = fa
        self._sms: float = sms
        self._sds: float = sds
        self._sdcs: str = sdcs
        self._s1rt: float = s1rt
        self._cr1: float = cr1
        self._s1uh: float = s1uh
        self._s1d: float = s1d
        self._s1: float = s1
        self._fv: float = fv
        self._sm1: float = sm1
        self._sd1: float = sd1
        self._sdc1: str = sdc1
        self._sdc: str = sdc
        self._tl: float = tl
        self._cv: float = cv
        self._twoPeriodDesignSpectrum = np.asarray(twoPeriodDesignSpectrum)
        self._twoPeriodMCErSpectrum = np.asarray(twoPeriodMCErSpectrum)
        self._verticalDesignSpectrum = np.asarray(verticalDesignSpectrum)
        self._verticalMCErSpectrum = np.asarray(verticalMCErSpectrum)

    @property
    def pgauh(self):
        """Uniform-hazard (2% probability of exceedance in 50 years) peak ground
        acceleration"""
        return self._pgauh

    @property
    def pgad(self):
        """Factored deterministic acceleration value (peak ground acceleration)"""
        return self._pgad

    @property
    def pga(self):
        """Maximum considered earthquake geometric mean (MCE_G) peak ground
        acceleration"""
        return self._pga

    @property
    def fpga(self):
        """Site amplification factor at peak ground acceleration"""
        return self._fpga

    @property
    def pgam(self):
        """Site-modified peak ground acceleration"""
        return self._pgam

    @property
    def ssrt(self):
        """Probabilistic risk-targeted ground motion (0.2 s)"""
        return self._ssrt

    @property
    def crs(self):
        """Mapped risk coefficient value (0.2 s)"""
        return self._crs

    @property
    def ssuh(self):
        """Factored uniform-hazard (2% probability of exceedance in 50 years)
        spectral acceleration (0.2 s)"""
        return self._ssuh

    @property
    def ssd(self):
        """Factored deterministic acceleration value (0.2 s)"""
        return self._ssd

    @property
    def ss(self):
        """Risk-targeted maximum considered earthquake (MCE_R) ground motion
        response acceleration (0.2 s)"""
        return self._ss

    @property
    def fa(self):
        """Site amplification factor (0.2 s)"""
        return self._fa

    @property
    def sms(self):
        """Site-modified spectral acceleration value (0.2 s)"""
        return self._sms

    @property
    def sds(self):
        """Numeric seismic design value (0.2 s)"""
        return self._sds

    @property
    def sdcs(self):
        """Seismic design category from SDS"""
        return self._sdcs

    @property
    def s1rt(self):
        """Probabilistic risk-targeted ground motion (1.0 s)"""
        return self._s1rt

    @property
    def cr1(self):
        """Mapped risk coefficient value (1.0 s)"""
        return self._cr1

    @property
    def s1uh(self):
        """Factored uniform-hazard (2% probability of exceedance in 50 years
        spectral acceleration (1.0 s)"""
        return self._s1uh

    @property
    def s1d(self):
        """Factored deterministic acceleration value (1.0 s)"""
        return self._s1d

    @property
    def s1(self):
        """Risk-targeted maximum considered earthquake (MCE_R) ground motion
        response acceleration (1.0 s)"""
        return self._s1

    @property
    def fv(self):
        """Site amplification factor (1.0 s)"""
        return self._fv

    @property
    def sm1(self):
        """Site-modified spectral acceleration value (1.0 s)"""
        return self._sm1

    @property
    def sd1(self):
        """Numeric seismic design value (1.0 s)"""
        return self._sd1

    @property
    def sdc1(self):
        """Seismic design category from SD1"""
        return self._sdc1

    @property
    def sdc(self):
        """Seismic design category from sdcs and sdc1"""
        return self._sdc

    @property
    def tl(self):
        """Long-period transition period in seconds"""
        return self._tl

    @property
    def cv(self):
        """Vertical coefficient"""
        return self._cv

    @property
    def twoPeriodDesignSpectrum(self):
        """Design horizontal response spectrum"""
        return self._twoPeriodDesignSpectrum

    @property
    def twoPeriodMCErSpectrum(self):
        """Risk-targeted maximum considered earthquake (MCE_R) horizontal
        response spectrum"""
        return self._twoPeriodMCErSpectrum

    @property
    def verticalDesignSpectrum(self):
        """Design vertical response spectrum"""
        return self._verticalDesignSpectrum

    @property
    def verticalMCErSpectrum(self):
        """Risk-targeted maximum considered earthquake (MCE_R) vertical response
        spectrum"""
        return self._verticalMCErSpectrum

    def seismic_response_coeff(self, R, T):
        """Calculate the seismic response coefficient C_s for a building at this
        site.

        Parameters
        ----------
        R : float
            Response modification factor
        T : float
            Building fundamental period (seconds)
        """
        return seismic_response_coeff(R, 1.0, self.sds, self.sd1, self.s1, T, self.tl)

    @classmethod
    def from_usgs(cls, latitude, longitude, risk_category, site_class):
        """Get site-specific parameters from the USGS web service.

        Parameters
        ----------
        latitude : int
            Latitude of the site
        longitude : int
            Longitude of the site
        risk_category : str
            Risk category for the structure (e.g. 'II')
        site_class : str
            NEHRP site classification (e.g. 'D')
        """
        response = requests.get(
            'https://earthquake.usgs.gov/ws/designmaps/asce7-16.json',
            params={
                'latitude': latitude,
                'longitude': longitude,
                'riskCategory': risk_category,
                'siteClass': site_class,
                'title': '',  # I'm not sure what this is other than metadata
            })
        if response.status_code != 200:
            if response.status_code == 500:
                error_desc = json.loads(response.content)['response']
                raise RuntimeError(f'Server returned error: {error_desc!r}')
            else:
                raise RuntimeError(f'Server returned error code {response.status_code}')

        data = json.loads(response.content)['response']['data']
        for key in ['fpga_note', 'fv_note', 't-sub-l']:
            try:
                del data[key]
            except KeyError:
                pass
        return cls(**data)



"""Functions that format strings into the appropriate LaTeX format."""


def diff(f, x, n=1):
    """Derivative representation.
    
    Parameters
    ----------
    f:
        function to be differentiated.
    x:
        variable to diff w.r.t.
    n:
        number of times to diff. (default: 1)
    """
    differ = Rf"\mathrm{{d}}"
    if n > 1:
        level = f"^{n}"
    else:
        level = ''

    return Rf"\frac{{{differ}{level}{{{f}}}}}{{{differ}{{{x}}}{level}}}"



import dataclasses
import enum
import fractions
import importlib.resources
from typing import NamedTuple, Union
import warnings

import numpy as np
import pandas as pd
import unyt

from . import units

#===============================================================================
# Constants
#===============================================================================
TRUE_VALUES = ['T']
FALSE_VALUES = ['F']
NA_VALUES = ['\N{EN DASH}']


class SteelError(Exception):
    """Steel design errors."""
    pass


#===============================================================================
# Materials
#===============================================================================
@dataclasses.dataclass
class SteelMaterial():
    """A steel material.

    Parameters
    ----------
    name : str
        Name of the material.
    E : float, unyt.unyt_array
        Elastic modulus. If units are not specified, assumed to be ksi.
    Fy : float, unyt.unyt_array
        Design yield strength. If units are not specified, assumed to be ksi.
    Fu : float, unyt.unyt_array
        Design tensile strength. If units are not specified, assumed to be ksi.
    Ry : float
        Expected yield strength factor. Dimensionless.
    Rt : float
        Expected tensile strength factor. Dimensionless.
    """
    name: str
    E: unyt.unyt_quantity
    Fy: unyt.unyt_quantity
    Fu: unyt.unyt_quantity
    Ry: float
    Rt: float

    def __post_init__(self):
        self.E = units.process_unit_input(self.E, default_units='ksi')
        self.Fy = units.process_unit_input(self.Fy, default_units='ksi')
        self.Fu = units.process_unit_input(self.Fu, default_units='ksi')
        self.Ry = units.process_unit_input(self.Ry,
                                           default_units='dimensionless',
                                           convert=True).item()
        self.Rt = units.process_unit_input(self.Rt,
                                           default_units='dimensionless',
                                           convert=True).item()
        if self.Fy > self.Fu:
            raise SteelError('SteelMaterial: yield strength must'
                             ' be less than tensile strength')

    @property
    def eFy(self):
        return self.Fy*self.Ry

    @property
    def eFu(self):
        return self.Fu*self.Rt

    @classmethod
    def from_name(cls, name: str, grade: str = None, application: str = None):
        """Look up a steel material based on name, grade, and application.

        All look-ups are case-insensitive.

        Parameters
        ----------
        name : str
            The name of the material specification (e.g. 'A572').
        grade : int, str, optional
            The grade of the material (e.g. 50 or 'B'). Only required if
            multiple grades are available.
        application : {'brb', 'hot-rolled', 'hss', 'plate', 'rebar'}, optional
            The application the material is used for. Only required if multiple
            applications are defined for the same material specification.

        Raises
        ------
        ValueError
            If multiple materials match the provided information. For example,
            requesting an A500 steel also requires specifying a grade (A or B).
        """
        name, grade = _check_deprecated_material(name, grade)

        def normalize(input):
            if pd.isna(input):
                return slice(None)
            else:
                return str(input).casefold()

        material = cls._get_materials_db().loc[normalize(name),
                                               normalize(grade),
                                               normalize(application)]

        # Lookup succeeds if we get a Series (exact indexes) or if we get a
        # DataFrame of length 1 (one or more indexes were sliced, but still only
        # one result returned).
        if isinstance(material, pd.DataFrame):
            if len(material) != 1:
                raise ValueError('Multiple materials found: specify grade '
                                 'and/or application to narrow search')
            material = material.iloc[0]

        name, grade, application = material.name
        display_grade = '' if pd.isna(grade) else f' Gr. {grade}'
        display_name = f'{name}{display_grade} ({application})'
        return cls(display_name.title(), **material)

    @classmethod
    def available_materials(cls):
        """Return a DataFrame of the available materials, whose rows can be used
        as lookups for `from_name`."""
        return cls._get_materials_db().index.to_frame(index=False)

    @classmethod
    def _get_materials_db(cls) -> pd.DataFrame:
        # Delay loading the materials database until required.
        try:
            return cls._materials_db
        except AttributeError:
            cls._materials_db = _load_materials_db('steel-materials-us.csv')
            return cls._materials_db


def _check_deprecated_material(name, grade):
    if ' Gr. ' in name:
        _name, grade = name.split(' Gr. ')
        warnings.warn(f'Convert old material name {name!r} to '
                      f'({_name!r}, grade={grade!r})')
        return (_name, grade)
    else:
        return (name, grade)


def _load_materials_db(filename):
    with importlib.resources.path('libtalley', filename) as p:
        material_df = pd.read_csv(p, index_col=['name', 'grade', 'application'])
    material_df.sort_index(inplace=True)
    return material_df


#===============================================================================
# Shapes table
#===============================================================================
class ShapesTable():
    def __init__(self, data: pd.DataFrame, units: pd.Series):
        """
        Parameters
        ----------
        data : pd.DataFrame
            Base data for the table.
        units : pd.Series
            Units for each (numeric) column in `data`.
        """
        self.data = data
        self.units = units

    def get_prop(self, shape: str, prop: str):
        """Return a property from the table with units.

        If a property is not defined for the given shape, nan is returned. If
        units are not defined for the property, the raw quantity is returned
        from the table.

        Parameters
        ----------
        shape : str
            Name of the shape to look up.
        prop : str
            Name of the property to look up.

        Returns
        -------
        q : unyt.unyt_quantity
            Value of the property with units.

        Raises
        ------
        KeyError
            If `shape` is not found in the table; if `prop` is not found in the
            table.
        """
        value = self.data.at[shape.casefold(), prop]
        units = self.units.get(prop)
        if units is None:
            return value
        else:
            return unyt.unyt_quantity(value, units)

    def get_shape(self, shape: str, include_units: bool = True):
        """
        Parameters
        ----------
        shape : str
            Name of the shape to retrieve.
        include_units : bool
            Whether to include units in the returned shape data. (default: True)

        Returns
        -------
        pd.Series
        """
        shape_data = self.data.loc[shape.casefold()].dropna()
        if include_units:
            for prop, value in shape_data.items():
                units = self.units.get(prop)
                if units is not None:
                    shape_data[prop] = unyt.unyt_quantity(value, units)
        return shape_data

    def lightest_shape(self, shape_list):
        """Return the lightest shape (force/length) from the given list.

        Works across different shape series, e.g. comparing an HSS and W works
        fine. If two or more shapes have the same lightest weight, a shape is
        returned, but which one is returned is undefined.

        Parameters
        ----------
        shape_list : list
            List of shapes to check.

        Examples
        --------
        >>> lightest_shape(['W14X82', 'W44X335'])
        'W14X82'
        >>> lightest_shape(['W14X82', 'HSS4X4X1/2'])
        'HSS4X4X1/2'
        """
        return self.data.loc[pd.Series(shape_list).str.casefold()].W.idxmin()

    @classmethod
    def from_file(cls,
                  file,
                  units,
                  true_values=TRUE_VALUES,
                  false_values=FALSE_VALUES,
                  na_values=NA_VALUES):
        """Load a shapes table from a file.

        Parameters
        ----------
        file : str, file-like
            Name of the file to load.
        units : dict
            Dictionary of units, with keys corresponding to the column names.
        true_values : list, optional
            List of values to convert to ``True``. (default: ['T'])
        false_values : list, optional
            List of values to convert to ``False``. (default: ['F'])
        na_values : list, optional
            List of values to convert to ``nan``. (default: ['–']) (note that
            this is an en-dash U+2013, not an ASCII hyphen U+002D)
        """
        data = pd.read_csv(file,
                           true_values=true_values,
                           false_values=false_values,
                           na_values=na_values)
        data.index = pd.Index(data['AISC_Manual_Label'].str.casefold(), name='')

        # Convert fractions to floats
        def str2frac2float(s):
            return float(sum(fractions.Fraction(i) for i in s.split()))

        for column in data.columns[data.dtypes == object]:
            if column not in ['AISC_Manual_Label', 'Type']:
                notnull_data = data[column][data[column].notnull()]
                converted_data = notnull_data.apply(str2frac2float)
                data[column].update(converted_data)
                data[column] = pd.to_numeric(data[column])

        data.sort_index(inplace=True)
        units = pd.Series(units)
        units.sort_index(inplace=True)
        return cls(data, units)


_SHAPES_US_FILE = 'aisc-shapes-database-v15-0-US.csv.bz2'
_SHAPES_US_UNITS = {
    'W': 'lbf/ft',
    'A': 'inch**2',
    'd': 'inch',
    'ddet': 'inch',
    'Ht': 'inch',
    'h': 'inch',
    'OD': 'inch',
    'bf': 'inch',
    'bfdet': 'inch',
    'B': 'inch',
    'b': 'inch',
    'ID': 'inch',
    'tw': 'inch',
    'twdet': 'inch',
    'twdet/2': 'inch',
    'tf': 'inch',
    'tfdet': 'inch',
    't': 'inch',
    'tnom': 'inch',
    'tdes': 'inch',
    'kdes': 'inch',
    'kdet': 'inch',
    'k1': 'inch',
    'x': 'inch',
    'y': 'inch',
    'eo': 'inch',
    'xp': 'inch',
    'yp': 'inch',
    'bf/2tf': 'dimensionless',
    'b/t': 'dimensionless',
    'b/tdes': 'dimensionless',
    'h/tw': 'dimensionless',
    'h/tdes': 'dimensionless',
    'D/t': 'dimensionless',
    'Ix': 'inch**4',
    'Zx': 'inch**3',
    'Sx': 'inch**3',
    'rx': 'inch',
    'Iy': 'inch**4',
    'Zy': 'inch**3',
    'Sy': 'inch**3',
    'ry': 'inch',
    'Iz': 'inch**4',
    'rz': 'inch',
    'Sz': 'inch**3',
    'J': 'inch**4',
    'Cw': 'inch**6',
    'C': 'inch**3',
    'Wno': 'inch**2',
    'Sw1': 'inch**4',
    'Sw2': 'inch**4',
    'Sw3': 'inch**4',
    'Qf': 'inch**3',
    'Qw': 'inch**3',
    'ro': 'inch',
    'H': 'dimensionless',
    'tan(α)': 'dimensionless',
    'Iw': 'inch**4',
    'zA': 'inch',
    'zB': 'inch',
    'zC': 'inch',
    'wA': 'inch',
    'wB': 'inch',
    'wC': 'inch',
    'SzA': 'inch**3',
    'SzB': 'inch**3',
    'SzC': 'inch**3',
    'SwA': 'inch**3',
    'SwB': 'inch**3',
    'SwC': 'inch**3',
    'rts': 'inch',
    'ho': 'inch',
    'PA': 'inch',
    'PA2': 'inch',
    'PB': 'inch',
    'PC': 'inch',
    'PD': 'inch',
    'T': 'inch',
    'WGi': 'inch',
    'WGo': 'inch',
}

_SHAPES_SI_FILE = 'aisc-shapes-database-v15-0-SI.csv.bz2'
_SHAPES_SI_UNITS = {
    'W': 'kg/m',
    'A': 'mm**2',
    'd': 'mm',
    'ddet': 'mm',
    'Ht': 'mm',
    'h': 'mm',
    'OD': 'mm',
    'bf': 'mm',
    'bfdet': 'mm',
    'B': 'mm',
    'b': 'mm',
    'ID': 'mm',
    'tw': 'mm',
    'twdet': 'mm',
    'twdet/2': 'mm',
    'tf': 'mm',
    'tfdet': 'mm',
    't': 'mm',
    'tnom': 'mm',
    'tdes': 'mm',
    'kdes': 'mm',
    'kdet': 'mm',
    'k1': 'mm',
    'x': 'mm',
    'y': 'mm',
    'eo': 'mm',
    'xp': 'mm',
    'yp': 'mm',
    'bf/2tf': 'dimensionless',
    'b/t': 'dimensionless',
    'b/tdes': 'dimensionless',
    'h/tw': 'dimensionless',
    'h/tdes': 'dimensionless',
    'D/t': 'dimensionless',
    'Ix': '1e6*mm**4',
    'Zx': '1e3*mm**3',
    'Sx': '1e3*mm**3',
    'rx': 'mm',
    'Iy': '1e6*mm**4',
    'Zy': '1e3*mm**3',
    'Sy': '1e3*mm**3',
    'ry': 'mm',
    'Iz': '1e6*mm**4',
    'rz': 'mm',
    'Sz': '1e3*mm**3',
    'J': '1e3*mm**4',
    'Cw': '1e9*mm**6',
    'C': '1e3*mm**3',
    'Wno': 'mm**2',
    'Sw1': '1e6*mm**4',
    'Sw2': '1e6*mm**4',
    'Sw3': '1e6*mm**4',
    'Qf': '1e3*mm**3',
    'Qw': '1e3*mm**3',
    'ro': 'mm',
    'H': 'dimensionless',
    'tan(α)': 'dimensionless',
    'Iw': '1e6*mm**4',
    'zA': 'mm',
    'zB': 'mm',
    'zC': 'mm',
    'wA': 'mm',
    'wB': 'mm',
    'wC': 'mm',
    'SzA': '1e3*mm**3',
    'SzB': '1e3*mm**3',
    'SzC': '1e3*mm**3',
    'SwA': '1e3*mm**3',
    'SwB': '1e3*mm**3',
    'SwC': '1e3*mm**3',
    'rts': 'mm',
    'ho': 'mm',
    'PA': 'mm',
    'PA2': 'mm',
    'PB': 'mm',
    'PC': 'mm',
    'PD': 'mm',
    'T': 'mm',
    'WGi': 'mm',
    'WGo': 'mm',
}


def _load_shapes_db(filename, units):
    with importlib.resources.path('libtalley', filename) as p:
        return ShapesTable.from_file(p, units)


shapes_US = _load_shapes_db(_SHAPES_US_FILE, _SHAPES_US_UNITS)
shapes_SI = _load_shapes_db(_SHAPES_SI_FILE, _SHAPES_SI_UNITS)


def property_lookup(shape, prop):
    """Retrieve a property from the US shapes table.

    Returns values without units for legacy reasons.

    Parameters
    ----------
    shape : str
        Name of the shape to look up.
    prop : str
        Name of the property to look up.
    """
    return shapes_US.data.loc[shape][prop]


def lightest_shape(shape_list):
    """Return the lightest shape (force/length) from the given list.

    Works across different shape series, e.g. comparing an HSS and W works fine.
    If two or more shapes have the same lightest weight, a shape is returned,
    but which is one is undefined.

    Parameters
    ----------
    shape_list : list
        List of shapes to check.

    Examples
    --------
    >>> lightest_shape(['W14X82', 'W44X335'])
    'W14X82'
    >>> lightest_shape(['W14X82', 'HSS4X4X1/2'])
    'HSS4X4X1/2'
    """
    return shapes_US.lightest_shape(shape_list)


#===============================================================================
# Design
#===============================================================================
class MemberType(enum.Enum):
    BRACE = 'BRACE'
    BEAM = 'BEAM'
    COLUMN = 'COLUMN'


class Ductility(enum.Enum):
    HIGH = 'HIGH'
    MODERATE = 'MODERATE'


class WtrResults(NamedTuple):
    passed: bool
    ht: float
    ht_max: float
    bt: float
    bt_max: float


def check_seismic_wtr_wide_flange(shape: str,
                                  mem_type: Union[str, MemberType],
                                  level: Union[str, Ductility],
                                  Ca: float,
                                  material: SteelMaterial = None) -> WtrResults:
    """Check the width-to-thickness ratio of a W shape for the given ductility.

    Parameters
    ----------
    shape : str
        AISC manual name for the shape being checked.
    mem_type : MemberType
        MemberType of the member.
    level : Ductility
        Level of Ductility being checked.
    Ca : float
        = P_u / (φ_c * P_y); adjusts maximum web width-to-thickness ratio
        for beams and columns. Does not affect braces. Should be < 1.0.
    material : SteelMaterial, optional
        Material to use (default A992, Fy = 50 ksi)

    Returns
    -------
    passed:
        Bool pass/fail. (ht <= ht_max and bt <= bt_max)
    ht:
        The h/tw value for the section
    ht_max:
        The maximum h/tw value for the section
    bt:
        The bf/2tf value for the section
    bt_max:
        The maximum bf/2tf value for the section

    Reference
    ---------
    AISC 341-16, Table D1.1 (pp. 9.1-14 -- 9.1-17)
    """
    if isinstance(mem_type, MemberType):
        mem_type: str = mem_type.value
    if isinstance(level, Ductility):
        level: str = level.value
    wtr_max = {
        ('BRACE', 'MODERATE'): _wtr_brace,
        ('BRACE', 'HIGH'): _wtr_brace,
        ('COLUMN', 'MODERATE'): _wtr_beam_column_moderate,
        ('COLUMN', 'HIGH'): _wtr_beam_column_high,
        ('BEAM', 'MODERATE'): _wtr_beam_column_moderate,
        ('BEAM', 'HIGH'): _wtr_beam_column_high,
    }[mem_type, level]

    if material is None:
        material = SteelMaterial.from_name('A992')
    E_eFy = np.sqrt(material.E/material.eFy).to_value('dimensionless')

    ht = property_lookup(shape, 'h/tw')
    bt = property_lookup(shape, 'bf/2tf')
    ht_max, bt_max = wtr_max(E_eFy, Ca)

    return WtrResults(ht <= ht_max and bt <= bt_max, ht, ht_max, bt, bt_max)


def _wtr_brace(E_eFy):
    """Maximum width-to-thickness ratio for a brace."""
    ht_max = 1.57*E_eFy
    bt_max = ht_max
    return ht_max, bt_max


def _wtr_beam_column_moderate(E_eFy, Ca):
    """Maximum width-to-thickness ratio for a moderately ductile beam/column."""
    bt_max = 0.40*E_eFy
    if Ca <= 0.114:
        ht_max = 3.96*E_eFy*(1 - 3.04*Ca)
    else:
        ht_max = max(1.29*E_eFy*(2.12 - Ca), 1.57*E_eFy)
    return ht_max, bt_max


def _wtr_beam_column_high(E_eFy, Ca):
    """Maximum width-to-thickness ratio for a highly ductile beam/column."""
    bt_max = 0.32*E_eFy
    if Ca <= 0.114:
        ht_max = 2.57*E_eFy*(1 - 1.04*Ca)
    else:
        ht_max = max(0.88*E_eFy*(2.68 - Ca), 1.57*E_eFy)
    return ht_max, bt_max


class Capacity(NamedTuple):
    tension: unyt.unyt_quantity
    compression: unyt.unyt_quantity
    postbuckling: unyt.unyt_quantity


def brace_capacity(shape: str, length: unyt.unyt_quantity,
                   material: SteelMaterial) -> Capacity:
    """
    Parameters
    ----------
    shape : str
        Steel shape of the brace.
    length : unyt_quantity
        Unbraced length of the brace.
    material : SteelMaterial
        Brace material.
    """
    shape = shapes_US.get_shape(shape)
    ry = shape['ry']
    Fe = np.pi**2*material.E/(length/ry)**2
    RyFy = material.eFy
    RyFy_Fe = RyFy/Fe

    if RyFy_Fe <= 2.25:
        Fcre = 0.658**RyFy_Fe*RyFy
    else:
        Fcre = 0.877*Fe

    Ag = shape['A']
    tension = RyFy*Ag
    compression = min(tension, 1/0.877*Fcre*Ag)
    postbuckling = 0.3*compression

    return Capacity(tension, compression, postbuckling)
